/*
 *  Copyright 2010 argius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
package net.argius.stew.ui.window;

import static java.util.Collections.*;
import static net.argius.stew.Iteration.*;
import static net.argius.stew.ui.window.Resource.getString;
import static net.argius.stew.ui.window.WindowOutputProcessor.showInformationMessageDialog;

import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.event.*;
import java.sql.*;
import java.util.*;
import java.util.List;
import java.util.Map.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;

import net.argius.logging.*;
import net.argius.stew.*;

/**
 * f[^x[X\c[B
 */
final class DatabaseInfoTree extends JTree {

    private static final Logger log = LoggerFactory.getLogger(DatabaseInfoTree.class);

    private Connector currentConnector;
    private DatabaseMetaData dbmeta;

    /**
     * RXgN^B
     */
    DatabaseInfoTree() {
        setRootVisible(false);
        setShowsRootHandles(false);
        setScrollsOnExpand(true);
        setCellRenderer(new Renderer());
        setModel(new DefaultTreeModel(null));
        // [Cxg]
        // ANVCxg
        ActionUtility actionUtility = ActionUtility.getInstance(this);
        final String keyCopyFullName = "copy-full-name";
        final String keyRefresh = "refresh";
        final String keyGenerateWherePhrase = "generate-where-phrase";
        final String keyGenerateSelectPhrase = "generate-select-phrase";
        final String keyGenerateUpdateStatement = "generate-update-statement";
        final String keyGenerateInsertStatement = "generate-insert-statement";
        final Component parent = this;
        actionUtility.bindAction(new AbstractAction() {

            public void actionPerformed(ActionEvent e) {
                TreePath[] paths = getSelectionPaths();
                if (paths == null) {
                    return;
                }
                List<String> names = new ArrayList<String>(paths.length);
                for (TreePath path : paths) {
                    InfoNode node = (InfoNode)path.getLastPathComponent();
                    names.add(node.getNodeFullName());
                }
                String s = join(names, String.format("%n"));
                Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
                StringSelection sselection = new StringSelection(s);
                clipboard.setContents(sselection, sselection);
            }

        }, keyCopyFullName);
        actionUtility.bindAction(new AbstractAction(keyRefresh) {

            public void actionPerformed(ActionEvent e) {
                TreePath[] paths = getSelectionPaths();
                if (paths == null) {
                    return;
                }
                for (TreePath path : paths) {
                    refresh((InfoNode)path.getLastPathComponent());
                }
            }

        });
        actionUtility.bindAction(new AbstractAction(keyGenerateWherePhrase) {

            public void actionPerformed(ActionEvent e) {
                TreePath[] paths = getSelectionPaths();
                if (paths == null) {
                    return;
                }
                List<ColumnNode> columns = new ArrayList<ColumnNode>();
                Set<String> tableNames = collectTableName(paths, columns);
                if (tableNames.isEmpty()) {
                    return;
                }
                fireActionPerformed(generateEquivalentJoinClause(columns));
            }

        });
        actionUtility.bindAction(new AbstractAction(keyGenerateSelectPhrase) {

            public void actionPerformed(ActionEvent e) {
                TreePath[] paths = getSelectionPaths();
                if (paths == null) {
                    return;
                }
                List<ColumnNode> columns = new ArrayList<ColumnNode>();
                Set<String> tableNames = collectTableName(paths, columns);
                if (tableNames.isEmpty()) {
                    return;
                }
                List<String> columnNames = new ArrayList<String>();
                final boolean one = tableNames.size() == 1;
                for (ColumnNode node : columns) {
                    columnNames.add(one ? node.getName() : node.getNodeFullName());
                }
                final String columnString = (columnNames.isEmpty()) ? "*" : join(columnNames, ", ");
                final String tableString = join(tableNames, ", ");
                fireActionPerformed(String.format("SELECT %s FROM %s WHERE ",
                                                  columnString,
                                                  tableString));
            }

        });
        actionUtility.bindAction(new AbstractAction(keyGenerateUpdateStatement) {

            public void actionPerformed(ActionEvent e) {
                TreePath[] paths = getSelectionPaths();
                if (paths == null) {
                    return;
                }
                List<ColumnNode> columns = new ArrayList<ColumnNode>();
                Set<String> tableNames = collectTableName(paths, columns);
                if (tableNames.isEmpty()) {
                    return;
                }
                if (tableNames.size() != 1 || columns.isEmpty()) {
                    showInformationMessageDialog(parent,
                                                 getString("e.enables-select-just-1-table"),
                                                 null);
                    return;
                }
                final String tableName = tableNames.toArray(new String[1])[0];
                List<String> columnExpressions = new ArrayList<String>();
                for (ColumnNode columnNode : columns) {
                    columnExpressions.add(columnNode.getName() + "=?");
                }
                fireActionPerformed(String.format("UPDATE %s SET %s WHERE ",
                                                  tableName,
                                                  join(columnExpressions, ", ")));
            }

        });
        actionUtility.bindAction(new AbstractAction(keyGenerateInsertStatement) {

            public void actionPerformed(ActionEvent e) {
                TreePath[] paths = getSelectionPaths();
                if (paths == null) {
                    return;
                }
                List<ColumnNode> columns = new ArrayList<ColumnNode>();
                Set<String> tableNames = collectTableName(paths, columns);
                if (tableNames.isEmpty()) {
                    return;
                }
                if (tableNames.size() != 1) {
                    showInformationMessageDialog(parent,
                                                 getString("e.enables-select-just-1-table"),
                                                 null);
                    return;
                }
                final String tableName = tableNames.toArray(new String[1])[0];
                List<String> columnNames = new ArrayList<String>();
                final Iterable<ColumnNode> columnNodes;
                if (columns.isEmpty()) {
                    final TreePath path = paths[0];
                    TableNode tableNode = (TableNode)path.getLastPathComponent();
                    if (tableNode.getChildCount() == 0) {
                        expandPath(path);
                    }
                    @SuppressWarnings("unchecked")
                    Iterable<ColumnNode> it = asIterable(tableNode.children());
                    columnNodes = it;
                } else {
                    columnNodes = columns;
                }
                for (ColumnNode node : columnNodes) {
                    columnNames.add(node.getName());
                }
                fireActionPerformed(String.format("INSERT INTO %s (%s) VALUES (%s);",
                                                  tableName,
                                                  join(columnNames, ", "),
                                                  join(nCopies(columnNames.size(), "?"), ", ")));
            }

        });
    }

    static Set<String> collectTableName(TreePath[] paths, List<ColumnNode> out) {
        Set<String> a = new LinkedHashSet<String>();
        for (TreePath path : paths) {
            InfoNode node = (InfoNode)path.getLastPathComponent();
            if (node instanceof TableNode) {
                TableNode tableNode = (TableNode)node;
                if (tableNode.isKindOfTable()) {
                    a.add(tableNode.getNodeFullName());
                }
            } else if (node instanceof ColumnNode) {
                ColumnNode columnNode = (ColumnNode)node;
                if (columnNode.getTableNode().isKindOfTable()) {
                    a.add(columnNode.getTableNode().getName());
                    out.add(columnNode);
                }
            }
        }
        return a;
    }

    void addActionListener(ActionListener listener) {
        listenerList.add(ActionListener.class, listener);
    }

    void removeActionListener(ActionListener listener) {
        listenerList.remove(ActionListener.class, listener);
    }

    void fireActionPerformed(String cmd) {
        ActionEvent event = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, cmd);
        for (ActionListener listener : listenerList.getListeners(ActionListener.class)) {
            listener.actionPerformed(event);
        }
    }

    /**
     * RootvfĐݒ肷B
     * @param env Environment
     * @throws SQLException
     */
    void refreshRoot(Environment env) throws SQLException {
        Connector c = env.getCurrentConnector();
        if (c == null) {
            if (log.isDebugEnabled()) {
                log.debug("not connected");
            }
            currentConnector = null;
            return;
        }
        if (c == currentConnector && getModel().getRoot() != null) {
            if (log.isDebugEnabled()) {
                log.debug("not changed");
            }
            return;
        }
        if (log.isDebugEnabled()) {
            log.debug("updating");
        }
        // f
        ConnectorNode connectorNode = new ConnectorNode(c.getName());
        DefaultTreeModel model = new DefaultTreeModel(connectorNode);
        setModel(model);
        final DefaultTreeSelectionModel m = new DefaultTreeSelectionModel();
        m.setSelectionMode(TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
        setSelectionModel(m);
        // m[h
        final DatabaseMetaData dbmeta = env.getCurrentConnection().getMetaData();
        final Set<InfoNode> createdStatusSet = new HashSet<InfoNode>();
        expandNode(connectorNode, dbmeta);
        createdStatusSet.add(connectorNode);
        expandPath(new TreePath(connectorNode.getFirstChild()));
        // Cxg
        addTreeWillExpandListener(new TreeWillExpandListener() {

            public void treeWillExpand(TreeExpansionEvent event) throws ExpandVetoException {
                TreePath path = event.getPath();
                final Object lastPathComponent = path.getLastPathComponent();
                if (!createdStatusSet.contains(lastPathComponent)) {
                    InfoNode node = (InfoNode)lastPathComponent;
                    createdStatusSet.add(node);
                    try {
                        expandNode(node, dbmeta);
                    } catch (SQLException ex) {
                        throw new RuntimeException(ex);
                    }
                }
            }

            public void treeWillCollapse(TreeExpansionEvent event) throws ExpandVetoException {
                // ignore
            }

        });
        this.dbmeta = dbmeta;
        // \
        model.reload();
        setRootVisible(true);
        this.currentConnector = c;
    }

    /**
     * m[hȉŐVԂɍXVB
     * @param node m[h
     */
    void refresh(InfoNode node) {
        if (dbmeta == null) {
            return;
        }
        node.removeAllChildren();
        final DefaultTreeModel model = (DefaultTreeModel)getModel();
        model.reload(node);
        try {
            expandNode(node, dbmeta);
        } catch (SQLException ex) {
            throw new RuntimeException(ex);
        }
    }

    /**
     * qm[hݒ肵ēWJB
     * @param parent em[h
     * @param dbmeta DB^
     * @throws SQLException
     */
    void expandNode(InfoNode parent, DatabaseMetaData dbmeta) throws SQLException {
        final DefaultTreeModel model = (DefaultTreeModel)getModel();
        for (InfoNode child : parent.createChildren(dbmeta)) {
            model.insertNodeInto(child, parent, parent.getChildCount());
        }
    }

    /**
     * NAB
     */
    void clear() {
        for (TreeWillExpandListener listener : getListeners(TreeWillExpandListener.class).clone()) {
            removeTreeWillExpandListener(listener);
        }
        setModel(new DefaultTreeModel(null));
        currentConnector = null;
        dbmeta = null;
        if (log.isDebugEnabled()) {
            log.debug("cleared");
        }
    }

    /**
     * 𐶐B
     * @param nodes m[hXg
     * @return 
     */
    static String generateEquivalentJoinClause(List<ColumnNode> nodes) {
        if (nodes.isEmpty()) {
            return "";
        }
        ListMap tm = new ListMap();
        ListMap cm = new ListMap();
        for (ColumnNode node : nodes) {
            final String tableName = node.getTableNode().getName();
            final String columnName = node.getName();
            tm.add(tableName, columnName);
            cm.add(columnName, String.format("%s.%s", tableName, columnName));
        }
        List<String> expressions = new ArrayList<String>();
        if (tm.size() == 1) {
            for (ColumnNode node : nodes) {
                expressions.add(String.format("%s=?", node.getName()));
            }
        } else {
            final String tableName = nodes.get(0).getTableNode().getName();
            for (String c : tm.get(tableName)) {
                expressions.add(String.format("%s.%s=?", tableName, c));
            }
            for (Entry<String, List<String>> entry : cm.entrySet()) {
                if (!entry.getKey().equals(tableName) && entry.getValue().size() == 1) {
                    expressions.add(String.format("%s=?", entry.getValue().get(0)));
                }
            }
            for (Entry<String, List<String>> entry : cm.entrySet()) {
                Object[] a = entry.getValue().toArray();
                final int n = a.length;
                for (int i = 0; i < n; i++) {
                    for (int j = i + 1; j < n; j++) {
                        expressions.add(String.format("%s=%s", a[i], a[j]));
                    }
                }
            }
        }
        return join(expressions, " AND ") + ';';
    }

    /**
     * Xg}bvB
     */
    private static final class ListMap extends LinkedHashMap<String, List<String>> {

        ListMap() {
            // empty
        }

        void add(String key, String value) {
            if (get(key) == null) {
                put(key, new ArrayList<String>());
            }
            get(key).add(value);
        }

    }

    /**
     * vf̕`揈(Renderer)B
     */
    private static class Renderer extends DefaultTreeCellRenderer {

        Renderer() {
            // empty
        }

        @Override
        public Component getTreeCellRendererComponent(JTree tree,
                                                      Object value,
                                                      boolean sel,
                                                      boolean expanded,
                                                      boolean leaf,
                                                      int row,
                                                      boolean hasFocus) {
            super.getTreeCellRendererComponent(tree, value, sel, expanded, leaf, row, hasFocus);
            if (value instanceof InfoNode) {
                setIcon(Resource.getImageIcon(((InfoNode)value).getIconName()));
            }
            return this;
        }

    }

    /**
     * m[hB
     */
    private abstract static class InfoNode extends DefaultMutableTreeNode {

        InfoNode(Object userObject) {
            super(userObject, true);
        }

        @Override
        public boolean isLeaf() {
            return false;
        }

        abstract protected List<InfoNode> createChildren(DatabaseMetaData dbmeta) throws SQLException;

        String getIconName() {
            final String className = getClass().getName();
            final String nodeType = className.replaceFirst(".+?([^\\$]+)Node$", "$1");
            return "node-" + nodeType.toLowerCase() + ".png";
        }

        protected String getNodeFullName() {
            return String.valueOf(userObject);
        }

    }

    /**
     * RlN^̃m[hB
     */
    private static class ConnectorNode extends InfoNode {

        ConnectorNode(String name) {
            super(name);
        }

        @Override
        protected List<InfoNode> createChildren(DatabaseMetaData dbmeta) throws SQLException {
            List<InfoNode> a = new ArrayList<InfoNode>();
            if (dbmeta.supportsCatalogsInDataManipulation()) {
                ResultSet rs = dbmeta.getCatalogs();
                try {
                    while (rs.next()) {
                        a.add(new CatalogNode(rs.getString(1)));
                    }
                } finally {
                    rs.close();
                }
            } else if (dbmeta.supportsSchemasInDataManipulation()) {
                ResultSet rs = dbmeta.getSchemas();
                try {
                    while (rs.next()) {
                        a.add(new SchemaNode(null, rs.getString(1)));
                    }
                } finally {
                    rs.close();
                }
            } else {
                ResultSet rs = dbmeta.getTableTypes();
                try {
                    while (rs.next()) {
                        TableTypeNode typeNode = new TableTypeNode(null, null, rs.getString(1));
                        if (typeNode.hasItems(dbmeta)) {
                            a.add(typeNode);
                        }
                    }
                } finally {
                    rs.close();
                }
            }
            return a;
        }

    }

    /**
     * J^Õm[hB
     */
    private static final class CatalogNode extends InfoNode {

        private final String name;

        CatalogNode(String name) {
            super(name);
            this.name = name;
        }

        @Override
        protected List<InfoNode> createChildren(DatabaseMetaData dbmeta) throws SQLException {
            List<InfoNode> a = new ArrayList<InfoNode>();
            if (dbmeta.supportsSchemasInDataManipulation()) {
                ResultSet rs = dbmeta.getSchemas();
                try {
                    while (rs.next()) {
                        a.add(new SchemaNode(name, rs.getString(1)));
                    }
                } finally {
                    rs.close();
                }
            } else {
                ResultSet rs = dbmeta.getTableTypes();
                try {
                    while (rs.next()) {
                        TableTypeNode typeNode = new TableTypeNode(name, null, rs.getString(1));
                        if (typeNode.hasItems(dbmeta)) {
                            a.add(typeNode);
                        }
                    }
                } finally {
                    rs.close();
                }
            }
            return a;
        }

    }

    /**
     * XL[}̃m[hB
     */
    private static final class SchemaNode extends InfoNode {

        private final String catalog;
        private final String schema;

        SchemaNode(String catalog, String schema) {
            super(schema);
            this.catalog = catalog;
            this.schema = schema;
        }

        @Override
        protected List<InfoNode> createChildren(DatabaseMetaData dbmeta) throws SQLException {
            List<InfoNode> a = new ArrayList<InfoNode>();
            ResultSet rs = dbmeta.getTableTypes();
            try {
                while (rs.next()) {
                    TableTypeNode typeNode = new TableTypeNode(catalog, schema, rs.getString(1));
                    if (typeNode.hasItems(dbmeta)) {
                        a.add(typeNode);
                    }
                }
            } finally {
                rs.close();
            }
            return a;
        }

    }

    /**
     * e[uʂ̃m[hB
     */
    private static final class TableTypeNode extends InfoNode {

        private static final String ICON_NAME_FORMAT = "node-tabletype-%s.png";

        private final String catalog;
        private final String schema;
        private final String tableType;

        TableTypeNode(String catalog, String schema, String tableType) {
            super(tableType);
            this.catalog = catalog;
            this.schema = schema;
            this.tableType = tableType;
        }

        @Override
        protected List<InfoNode> createChildren(DatabaseMetaData dbmeta) throws SQLException {
            List<InfoNode> a = new ArrayList<InfoNode>();
            ResultSet rs = dbmeta.getTables(catalog, schema, null, new String[]{tableType});
            try {
                while (rs.next()) {
                    final String table = rs.getString(3);
                    final String type = rs.getString(4);
                    final boolean kindOfTable = type.matches("TABLE|VIEW|SYNONIM");
                    a.add(new TableNode(catalog, schema, table, kindOfTable));
                }
            } finally {
                rs.close();
            }
            return a;
        }

        @Override
        String getIconName() {
            final String name = String.format(ICON_NAME_FORMAT, getUserObject());
            if (getClass().getResource("icon/" + name) == null) {
                return String.format(ICON_NAME_FORMAT, "");
            }
            return name;
        }

        /**
         * vf݂邩ǂׂB
         * @param dbmeta DB^
         * @return vf݂邩ǂ
         * @throws SQLException
         */
        boolean hasItems(DatabaseMetaData dbmeta) throws SQLException {
            ResultSet rs = dbmeta.getTables(catalog, schema, null, new String[]{tableType});
            try {
                return rs.next();
            } finally {
                rs.close();
            }
        }

    }

    /**
     * e[ũm[hB
     */
    static final class TableNode extends InfoNode {

        private final String catalog;
        private final String schema;
        private final String name;
        private final boolean kindOfTable;

        TableNode(String catalog, String schema, String name, boolean kindOfTable) {
            super(name);
            this.catalog = catalog;
            this.schema = schema;
            this.name = name;
            this.kindOfTable = kindOfTable;
        }

        @Override
        protected List<InfoNode> createChildren(DatabaseMetaData dbmeta) throws SQLException {
            List<InfoNode> a = new ArrayList<InfoNode>();
            ResultSet rs = dbmeta.getColumns(catalog, schema, name, null);
            try {
                while (rs.next()) {
                    a.add(new ColumnNode(rs.getString(4),
                                         rs.getString(6),
                                         rs.getInt(7),
                                         rs.getString(18),
                                         this));
                }
            } finally {
                rs.close();
            }
            return a;
        }

        @Override
        public boolean isLeaf() {
            return false;
        }

        @Override
        protected String getNodeFullName() {
            List<String> a = new ArrayList<String>();
            if (catalog != null) {
                a.add(catalog);
            }
            if (schema != null) {
                a.add(schema);
            }
            a.add(name);
            return Iteration.join(a, ".");
        }

        String getName() {
            return name;
        }

        boolean isKindOfTable() {
            return kindOfTable;
        }

    }

    /**
     * ̃m[hB
     */
    static final class ColumnNode extends InfoNode {

        private String name;
        private TableNode tableNode;

        ColumnNode(String name, String type, int size, String nulls, TableNode tableNode) {
            super(format(name, type, size, nulls));
            setAllowsChildren(false);
            this.name = name;
            this.tableNode = tableNode;
        }

        String getName() {
            return name;
        }

        TableNode getTableNode() {
            return tableNode;
        }

        private static String format(String name, String type, int size, String nulls) {
            final String nonNull = "NO".equals(nulls) ? " NOT NULL" : "";
            return String.format("%s [%s(%d)%s]", name, type, size, nonNull);
        }

        @Override
        public boolean isLeaf() {
            return true;
        }

        @Override
        protected List<InfoNode> createChildren(DatabaseMetaData dbmeta) throws SQLException {
            return emptyList();
        }

        @Override
        protected String getNodeFullName() {
            return String.format("%s.%s", tableNode.getName(), name);
        }

    }

}
